#coding:utf-8
"""
Created on 18.09.14

@author: kasimova
"""
import json
import datetime

from collections import defaultdict
from django.conf import settings as kinder_settings
from django.utils import timezone
from yadic.container import Injectable

from kinder.core.audit_log.models import AdditionalEncoder
from kinder.core.audit_log.helpers import get_model, get_field
from kinder.core.declaration.models import (DeclarationPrivilege,
                                            DeclarationUnit,
                                            Declaration,
                                            DeclarationStatus)
from kinder.core.declaration_status.models import DSS
from kinder.core.privilege.models import Privilege
from kinder.core.unit.models import Unit
from kinder.core.helpers import (recursive_getattr, recursive_setattr,
                                 recursive_get_verbose_name)
from concentrator.models import ChangeDeclaration, ChangeStatus
from concentrator.changes.helpers import *
from concentrator.changes import rules
from concentrator.rules import DeclarationStatusChangeRule
from concentrator.exceptions import ValidationError

from .models import UpdateParams, ChangeSource


TYPE_CHOICE = 'choices'


class ChangesMap(object):
    """Соответствие полей django-моделей и полей моделей сервиса
    с учетом их типа.
    """

    _map = {
        # TODO: удалить проверку тех полей, которых никто
        # Никогда не будет изменять на строне концетратора
        'Declaration': [
            ('desired_date', datetime.date, 'EntryDate'),
            ('offer_other', bool, 'EduOrganizations\AllowOfferOther'),
            ('comment_privileges', str, 'Benefits\BenefitsDocInfo'),
            ('work_type', TYPE_CHOICE, 'Schedule\ScheduleType'),
        ],
        'Delegate': [
            ('firstname', str, 'Applicant\FirstName'),
            ('surname', str, 'Applicant\LastName'),
            ('patronymic', str, 'Applicant\MiddleName'),
            ('snils', str, 'Applicant\Snils'),
            ('dul_type', TYPE_CHOICE, 'Applicant\DocType'),
            ('dul_series', str, 'Applicant\DocSeria'),
            ('dul_number', str, 'Applicant\DocNumber'),
            ('dul_issued_by', str, 'Applicant\DocIssuerName'),
            ('dul_place', str, 'Applicant\DocIssuerDepartmentCode'),
            ('dul_date', datetime.date, 'Applicant\DocIssueDate'),
            ('type', int, 'Applicant\ApplicantType'),
            ('email', str, 'Applicant\Email'),
            ('phones', str, 'Applicant\PhoneNumber'),
        ],
        'Children': [
            ('firstname', str, 'DeclaredPerson\FirstName'),
            ('surname', str, 'DeclaredPerson\LastName'),
            ('patronymic', str, 'DeclaredPerson\MiddleName'),
            ('snils', str, 'DeclaredPerson\Snils'),
            ('address_full', str, 'DeclaredPerson\AddressResidence'),
            ('reg_address_full', str,
             'DeclaredPerson\AddressRegistration'),
            ('dul_series', str, 'DeclaredPerson\BirthDocSeria'),
            ('dul_number', str, 'DeclaredPerson\BirthDocNumber'),
            ('dul_date', datetime.date,
             'DeclaredPerson\BirthDocIssueDate'),
            ('gender', int, 'DeclaredPerson\Sex'),
            ('health_need', TYPE_CHOICE, 'AdaptationProgramType'),
            ('health_need_confirmation', str, 'AdaptationProgramDocInfo'),
            ('zags_act_number', str, 'DeclaredPerson\BirthDocActNumber'),
            ('zags_act_place', str, 'DeclaredPerson\BirthDocIssuer'),
            ('zags_act_date', datetime.date, 'DateOfActNumber'),
            ('birthplace', str, 'DeclaredPerson\BirthPlace'),
        ],
        # Сейчас для применения изменений всегда используется
        # концентраторский ChangesMap, а для сервисных методов липецкий,
        # поэтому пришлось продублировать.
        'PrivilegeConfirmationAttributes': [
            ('privilege_owner', int, 'WhoHaveBenefit\Type'),
            ('delegate.firstname', str, 'WhoHaveBenefit\FirstName'),
            ('delegate.surname', str, 'WhoHaveBenefit\LastName'),
            ('delegate.patronymic', str, 'WhoHaveBenefit\MiddleName'),
            ('delegate.snils', str, 'WhoHaveBenefit\Snils'),
            ('delegate.dul_type', TYPE_CHOICE, 'WhoHaveBenefit\DocType'),
            ('delegate.dul_series', str, 'WhoHaveBenefit\DocSeria'),
            ('delegate.dul_number', str, 'WhoHaveBenefit\DocNumber'),
            ('delegate.dul_place', str, 'WhoHaveBenefit\DocIssuerDepartmentCode'),
            ('delegate.dul_issued_by', str, 'WhoHaveBenefit\DocIssuerName'),
            ('delegate.dul_date', datetime.date, 'WhoHaveBenefit\DocIssueDate'),
        ]
    }

    @property
    def model_name_list(self):
        u"""
        Список имён классов моделей.

        :rtype: list
        """
        return sorted(self._map.keys())

    def extend_changes_map(self, model_name, fields_description):
        u"""
        Расширение (плагинами) соответствия полей.

        :param str model_name: имя класса джанга-модели
        :param list fields_description: список кортежей из:
            - имя поля джанга модели,
            - тип поля,
            - имя поля модели веб-сервиса
        """
        assert isinstance(fields_description, list)
        self._map[model_name] = fields_description

    def get(self, model_name):
        u"""Получение данных об обрабатываемых полях по имени модели."""
        return self._map[model_name]

    def get_field_info(self, name_field, name_model):
        """
        отдает информацию по полю
        @return:
        """
        tmp_list = filter(lambda x: x[0] == name_field, self.get(name_model))
        if len(tmp_list) != 1:
            raise Exception(u"Поле %s не найдено в описании модели %s" %
                            (name_field, name_model))
        return tmp_list[0]

    def get_read_only_fields(self):
        """
        Возвращает список полей недоступных для редактирования в ЕПГУ
        Это все поля из списка self._map,
        и не указанные в справочнике доступных

        :rtype: list
        """
        result = []
        editable_list = UpdateParams.objects.all().values_list(
            'model_name', 'field_name')
        for model_name in self.model_name_list:
            for sys_name, type_value, conc_name in self.get(model_name):
                if (model_name, sys_name) in editable_list:
                    continue
                result.append(conc_name)

        return result

    def get_function_compare(self, type_value):
        u"""
        Возвращает функцию сравнения, в зависимости от типа данных.

        Если не включено USE_TZ джанга хранит Naive datetime objects,
        а из сервсов приходят aware datetime objects,
        для их сравнения необходимо приведение.

        :raise: Exception
        :rtype: callable
        :returns: функция сравнения
        """
        if type_value == TYPE_CHOICE:
            def compare(x, y):
                if x and y:
                    return x.id == y.id
                else:
                    return x == y
            return lambda x, y: compare(x, y)

        elif type_value is datetime.date:
            def compare_dates(x, y):
                new_x = x
                new_y = y
                if isinstance(x, datetime.datetime):
                    new_x = x.date()
                if isinstance(y, datetime.datetime):
                    new_y = y.date()
                return new_x == new_y
            return lambda x, y: compare_dates(x, y)

        elif type_value is str:
            def compare_str(x, y):
                new_x = x if x else ''
                new_y = y if y else ''
                return new_x.lower() == new_y.lower()
            return lambda x, y: compare_str(x, y)

        elif type_value in [bool, int]:
            return lambda x, y: x == y

        elif type_value is datetime.datetime:
            def compare_datetimes(x, y):
                new_x = (
                    timezone.make_naive(x, timezone.get_default_timezone())
                    if timezone.is_aware(x) else x
                )
                new_y = y
                return new_x == new_y
            if kinder_settings.USE_TZ:
                return lambda x, y: x == y
            else:
                return lambda x, y: compare_datetimes(x, y)
        else:
            raise Exception(u'Неизвестный тип поля')

    def get_store_models(self):
        u"""
        Возвращает список моделей для интерфейса
        "Параметры для изменения данных через ЕПГУ"

        :rtype: list
        """
        return [(model_name, get_model(model_name)._meta.verbose_name)
                for model_name in self.model_name_list]

    def get_fields(self):
        u"""
        Список полей в привязке к моделям для интерфейса
        "Параметры для изменения данных через ЕПГУ".

        :returns: словарь вида
            {'Declaration':[('date', 'Дата подачи'),
            ('desired_date', '')],...}
        :rtype: dict
        """
        result = defaultdict(list)
        for model_name, field_description in self._map.iteritems():
            model = get_model(model_name)
            list_fields = []
            for field_name, _, _ in field_description:
                field = get_field(model, field_name)
                if field and hasattr(field, 'verbose_name'):
                    verbose_name_field = field.verbose_name
                else:
                    continue
                list_fields.append((field_name, verbose_name_field))
            result[model_name] = list_fields
        return dict(result)


map_changes = ChangesMap()


class ChangeHelper(object):
    """
    Класс выполняет роль ЧЯ
    Принимает изменения, хранит, выводит и применяет их
    Скрывает механизм хранения и обработки
    """
    def __init__(self, model, name_model):
        self.model = model
        # из map_changes и rules.display_changes_map
        self.name_model = name_model
        self._result = []
        self.model_field = dict([(f.name, f) for f in model._meta.fields])

    def check_diff(self, conc_object, save_object):
        """
        Проверяет пришли ли изменения по объекту из концетратра
        @param conc_object: объект построенный по данным из концетратора
        @param save_object: объект из БД
        @return: сохраняет в self.result данные об изменениях, вида:
        """
        instanse_model = save_object or conc_object
        for field, type_value, _ in map_changes.get(self.name_model):
            conc_value = recursive_getattr(conc_object,
                                           field.replace('.', '__'))
            save_value = recursive_getattr(save_object,
                                           field.replace('.', '__'))
            compare_func = map_changes.get_function_compare(type_value)
            if not compare_func(conc_value, save_value):
                self._handling_change(field, type_value, conc_value,
                                      save_value, instanse_model)

    def _handling_change(self, field, type_value, conc_value, save_value,
                         instanse_model):
        """
        :param field: имя поля модели
        :param type_value: тип значения поля: int, str и тд
        :param conc_value: значение поля модели из концентратора
        :param save_value: текущее значение поля модели
        :param instanse_model: инстанс django модели ,
        для получения verbose_name поля
        @return:
        """
        if '.' in field:
            name_field = recursive_get_verbose_name(instanse_model,
                                                    field.replace('.', '__'))
        else:
            name_field = self.model_field[field].verbose_name

        if type_value == TYPE_CHOICE:
            self._result.append(
                {field: (name_field,
                         conc_value.id if conc_value else conc_value,
                         save_value.id if save_value else save_value)})
        else:
            self._result.append(
                {field: (name_field, conc_value, save_value)})

    def get_result(self):
        return self._result

    def is_related(self, data):
        """
        если есть model то это _set а не field
        @param data:
        @return:
        """
        return hasattr(data, 'model')

    def show_change(self, difference, changes):
        """
        Метод вернет информацию об изменении в человекопонятном виде
        :return: список из { имени поля, формат. старого значения,
                формат. нового значения }
        :rtype: list
        """
        rows = []
        get_data = rules.display_changes_map.get(self.name_model)
        # состояние модели на текущйи момент
        data_now = get_data(changes)

        for change_field in difference:
            name_field = change_field.keys()[0]
            verbosename_field, new_value, old_value = change_field.values()[0]
            #  тк есть зависмость от именеи полей в ChangesDetailRowsAction
            rows.append({
                'field': format_change(self.model._meta.verbose_name,
                                       verbosename_field),
                'old_value': convert_value(
                    rules.DISPLAY_TYPES_MAP, data_now,
                    name_field, old_value),
                'new_value': convert_value(
                    rules.DISPLAY_TYPES_MAP, data_now,
                    name_field, new_value)
            })

        return rows

    def apply_change(self, difference, changes):
        """
        - применяет изменения в системе
        - возаращает список изменений в окно
        @return:
        """
        updated_fields = {}
        list_reference = []

        get_data = rules.display_changes_map.get(self.name_model)
        # состояние модели на текущйи момент
        data_now = get_data(changes)
        for change_field in difference:
            postfix = ''
            name_field = change_field.keys()[0]
            verbosename_field, new_value, old_value = change_field.values()[0]
            new_value = convert_value(
                rules.FIELD_TYPES_MAP, data_now, name_field, new_value)
            # Если были изменения по fk, мы их должны запомнить,
            # тк их сохранять нужно отдельно
            update_obj = recursive_setattr(
                data_now, name_field.replace('.', '__'), new_value)
            if update_obj != data_now and update_obj not in list_reference:
                list_reference.append(update_obj)

            # для FK Нужно подправить имя поля
            field_info = map_changes.get_field_info(
                name_field, self.name_model)
            show_value = convert_value(rules.RESULT_TYPES_MAP, data_now,
                                       name_field, new_value)
            if field_info[1] == TYPE_CHOICE:
                postfix = '_id'
                show_value = show_value.id if show_value else ''

            updated_fields['{0}.{1}{2}'.format(
                self.name_model.lower(), name_field, postfix)] = show_value
        # Сохранем все измения по FK
        map(lambda x: x.save(), list_reference)
        # Затем сам объект
        if data_now:
            data_now.save()
        return updated_fields


class StorageHelper(object):
    """
    Отвечает за управлением изменений по разным моделям в целом
    Работает со списком изменений пришедших от концентратора,и в зависмости от
    модели применяет различные ChangeHelper
    """

    __metaclass__ = Injectable
    depends_on = ("source", )

    # карта соответствия Модель <-> ChangeHelper
    _map_change_helpers = {
        'Default': ChangeHelper
    }
    # переопределяемые обработчики изменений
    _map_override_change_helpers = {}

    @classmethod
    def register_change_helper(cls, model_name, change_helper):
        u"""
        Регистрация обработчика изменений модели во внутреннем кеше.

        :param str model_name: имя модели, подлежайшей изменению
        :param change_helper: обработчик изменений
        :type change_helper: ChangeHelper
        """
        assert issubclass(change_helper, ChangeHelper)
        cls._map_change_helpers[model_name] = change_helper

    @classmethod
    def register_override_change_helper(cls, model_name, change_helper):
        u"""
        Регистрация переопределенного обработчика изменений модели.

        :param str model_name: имя модели, подлежайшей изменению
        :param change_helper: обработчик изменений
        :type change_helper: ChangeHelper
        """
        assert issubclass(change_helper, ChangeHelper)
        cls._map_override_change_helpers[model_name] = change_helper

    @classmethod
    def get_change_helper(cls, model_name):
        u"""
        Получение ChangeHelper в зависимости от модели.

        :rtype: ChangeHelper
        """
        override = cls._map_override_change_helpers.get(model_name)
        if override:
            return override
        else:
            return cls._map_change_helpers.get(
                model_name,
                cls._map_change_helpers['Default']
            )

    @classmethod
    def create_change(cls, declaration, list_ch_help,
                      source=ChangeSource.NEW_APPLICATION,
                      case_number=None):
        """
        @param list_ch_help : Список инстансов классов,каждый из
        котрых хранит изменени по 1 модели
        """
        result = {}
        for ch_help in list_ch_help:
            assert isinstance(ch_help, ChangeHelper)
            diff = ch_help.get_result()
            if diff:
                result[ch_help.name_model] = diff
        if result:
            change = ChangeDeclaration(
                declaration=declaration,
                data=json.dumps(result, cls=AdditionalEncoder),
                source=source,
                case_number=case_number
            )
            change.save()
            return True
        return False

    @classmethod
    def get_change(cls, changes):
        """
        :param changes запись из ChangeDeclaration
        :return: Возвращает кортеж, первое значение которго говорит о том,
        нужно ли отклонить набор изменений, полученных с ЕПГУ. А это необходимо
        при наличии невалидных данных.
        Второе значение кортежа содержит данные об изменениях
        :rtype: (bool, list)
        """
        result = []
        for name_model, difference in json.loads(changes.data).iteritems():
            class_model = get_model(name_model)
            change_helper = cls.get_change_helper(name_model)
            ch_help = change_helper(class_model, name_model)
            rows = ch_help.show_change(difference, changes)
            result.extend(rows)
        return result

    @classmethod
    def apply_changes(cls, changes, request, comment):
        """
        Сохраняет изменений, меняет статус на Исполнено
        @param cls:
        @param changes:
        @param request:
        @param comment:
        @return:
        """
        updated_fields = {}
        for name_model, difference in json.loads(changes.data).iteritems():
            class_model = get_model(name_model)
            change_helper = cls.get_change_helper(name_model)
            ch_help = change_helper(class_model, name_model)
            updated_fields.update(ch_help.apply_change(difference, changes))
        changes.user = request.user
        changes.state = ChangeStatus.ACCEPT
        changes.commentary = comment
        changes.save()
        return updated_fields


class DeclarationChangeHelper(ChangeHelper):
    """
    Переопределяем, из за специфичной обработки
    смены статуса заявки
    """
    NAME_FIELD = 'status'

    def get_new_status(self):
        """
        @return:
        """
        return DeclarationStatus.objects.get(code=DSS.ARCHIVE)

    def get_status_change(self, declaration):
        status_verbose_name = Declaration.status.field.verbose_name
        return [
            {self.NAME_FIELD: (
                status_verbose_name,
                self.get_new_status().id,
                declaration.status.id)}
        ]

    def apply_change(self, difference, changes):
        """
        применит изменение в системе
        @return:
        """
        if not filter(lambda x: self.NAME_FIELD in x.keys(), difference):
            updated_fields = super(DeclarationChangeHelper,
                                   self).apply_change(difference, changes)
        else:
            new_status = self.get_new_status()
            DeclarationStatusChangeRule.change(changes.declaration, new_status)
            updated_fields = {'declaration.status_id': new_status.id}
        return updated_fields


class SetChangeHelper(ChangeHelper):
    NEW_NAME = 'conc_unit'
    SYS_NAME = 'sys_unit'


class DeclarationUnitChangeHelper(SetChangeHelper):
    """
    Изменения желаемых учреждений храним в другом формате,
    поэтому переопределяем мотоды сохранения, показа иприменения изменений
    """
    def check_diff(self, conc_object, save_object):
        """
        Проверяет пришли ли изменения по объекту из концетратра
        - если пришли сохраняем желаемое состояние
        и текущее состояния, для показа в дальнейшем истории изменений
        @param conc_object: список из (id, ord)
        жел. учреждений из запроса на изменение
        @param save_object: список из (id, ord) текущих
        жел. учреждений
        @return: сохраняет в self.result данные об изменениях, вида:
        [{"conc_unit": [4, 1]}, {"conc_unit": [24, 2]}, {"sys_unit": [4, 1]}]
        """
        set_conc = set(conc_object)
        set_sys = set(save_object)
        if set_conc ^ set_sys:
            diff = [{self.NEW_NAME: tuple(el)}
                    for el in conc_object]
            diff.extend([
                {self.SYS_NAME: tuple(el)} for el in save_object
            ])
            self._result.extend(diff)

    def _get_diff_data(self, curent_state, future_state):
        """
        Разврачивает изменения на 3 списка, на изменение, на удаление
        на добавление
        @param curent_state:
        @param future_state:
        @return: new_list, delete_list, update_list
        """
        new_list = []
        delete_list = list(curent_state)
        update_list = []
        # Проверяет что изменилось, удалилось и добавилось
        for change in future_state:
            unit_id, new_ord = change[self.NEW_NAME]
            find = False
            for du_id, du_ord in curent_state:
                if du_id == unit_id and du_ord != new_ord:
                    update_list.append((unit_id, new_ord, du_ord))
                    delete_list.remove((du_id, du_ord))
                    find = True
                    break
                elif du_id == unit_id:
                    delete_list.remove((du_id, du_ord))
                    find = True
                    break
            if not find:
                new_list.append((unit_id, new_ord, 0))
        # все словари должны содержать туплы из 3 значений
        delete_list = map(lambda x: (x[0], 0, x[1]), delete_list)
        return new_list, delete_list, update_list

    def _extend_show_result(self, get_field, get_old, get_new, list_changes,
                            rows):
        for uni_id, new_ord, old_ord in list_changes:
            try:
                unit = Unit.objects.get(id=uni_id)
                rows.append({
                    'field': get_field(unit),
                    'old_value': get_old(unit, old_ord),
                    'new_value': get_new(unit, new_ord),
                })
            except Unit.DoesNotExist:
                pass
        return rows

    def _get_future_state(self, difference):
        return filter(lambda x: self.NEW_NAME in x.keys(), difference)

    def _get_current_state(self, difference, changes):
        if changes.state == ChangeStatus.WAIT:
            # состояние модели на текущий момент
            get_data = rules.display_changes_map.get(self.name_model)
            data_now = get_data(changes)
            curent_state = list(data_now.all().values_list('unit_id', 'ord'))
        else:
            curent_state = map(lambda x: tuple(x[self.SYS_NAME]),
                               filter(lambda x: self.SYS_NAME in x.keys(),
                                      difference))
        return curent_state

    def show_change(self, difference, changes):
        """
        Метод вернет информацию об изменении в человекопонятном виде
        - перед изменениями покажет разницу с текущим состоянием БД
        - после изменения с состонияем на обработку сервиса
        :return: список из { имени поля, формат. старого значения,
                 формат. нового значения }
        :rtype:(bool, list)
        """
        rows = []
        verbose_name = self.model._meta.verbose_name
        curent_state = self._get_current_state(difference, changes)
        future_state = self._get_future_state(difference)
        new_list, delete_list, update_list = self._get_diff_data(
            curent_state, future_state)

        self._extend_show_result(
            lambda x: verbose_name,
            lambda x, y: '-',
            lambda x, y: x.display(),
            new_list,
            rows
        )
        self._extend_show_result(
            lambda x: verbose_name,
            lambda x, y: x.display(),
            lambda x, y: '-',
            delete_list,
            rows
        )
        self._extend_show_result(
            lambda x: format_change(
                '%s - %s' % (verbose_name, x.display()), u'Приоритет'),
            lambda x, y: y,
            lambda x, y: y,
            update_list,
            rows
        )
        return rows

    def apply_change(self, difference, changes):
        """
        применяит изменение в системе
        не возвращаем дикт с изменениями, тк он работает
        только для моделей заявки и ребенка
        @return:
        """
        curent_state = self._get_current_state(difference, changes)
        future_state = self._get_future_state(difference)
        new_list, delete_list, update_list = self._get_diff_data(
            curent_state, future_state)

        for unit_id, new_ord, old_ord in new_list:
            DeclarationUnit.objects.create(
                declaration=changes.declaration,
                unit_id=unit_id,
                ord=new_ord
            )
        for unit_id, new_ord, old_ord in update_list:
            du_qs = DeclarationUnit.objects.filter(
                declaration=changes.declaration,
                unit_id=unit_id,
            )
            if du_qs.count() == 1:
                du = du_qs[0]
                du.ord = new_ord
                du.save()

        for unit_id, _, _ in delete_list:
            du_qs = DeclarationUnit.objects.filter(
                declaration=changes.declaration,
                unit_id=unit_id,
            )
            if du_qs.count() == 1:
                du = du_qs[0]
                du.delete()
        return {}


class DeclarationPrivilegeChangeHelper(SetChangeHelper):

    def check_diff(self, conc_object, save_object):
        """
        Проверяет пришли ли изменения по объекту из концетратра
        - если пришли сохраняем желаемое состояние
        и текущее состояния, для показа в дальнейшем истории изменений
        @param conc_object: список из id
        жел. учреждений из запроса на изменение
        @param save_object: список из id текущих льгот
        @return: сохраняет в self.result данные об изменениях, вида:
            [{"conc_unit": [1]}, {"sys_unit": [1, 6]}]
        """
        set_conc = set(conc_object)
        set_sys = set(save_object)
        if set_conc ^ set_sys:
            self._result.extend(
                [{self.NEW_NAME: tuple(conc_object)},
                 {self.SYS_NAME: tuple(save_object)}])

    def _get_diff_data(self, curent_state, future_state):
        """
        Разврачивает изменения на 2 списка,  на удаление и на добавление
        @param curent_state:
        @param future_state:
        @return: new_list, delete_list
        """
        set_sys = set(curent_state)
        set_cur = set(future_state)
        new_list = list(set_cur - set_sys)
        delete_list = list(set_sys-set_cur)
        return new_list, delete_list

    def _extend_show_result(self, get_field, get_old, get_new, list_changes,
                            rows):
        for priv_id in list_changes:
            try:
                priv = Privilege.objects.get(id=priv_id)
                rows.append({
                    'field': get_field(priv),
                    'old_value': get_old(priv),
                    'new_value': get_new(priv),
                })
            except Privilege.DoesNotExist:
                pass
        return rows

    def _get_future_state(self, difference):
        """
        возвращает список льгот на добавление
        @param difference:
        @return:
        """
        future_state_l = map(lambda x: x[self.NEW_NAME],
                             filter(lambda x: self.NEW_NAME in x.keys(),
                                    difference))
        try:
            future_state = future_state_l[0]
        except IndexError:
            raise ValidationError
        return future_state

    def _get_current_state(self, difference, changes):
        """
        возвращает список id льгот заявки
        @param difference:
        @param changes:
        @return:
        """
        if changes.state == ChangeStatus.WAIT:
            # состояние модели на текущий момент
            get_data = rules.display_changes_map.get(self.name_model)
            data_now = get_data(changes)
            curent_state = list(data_now.all().values_list('privilege_id',
                                                           flat=True))
        else:
            curent_state = map(lambda x: x[self.SYS_NAME],
                               filter(lambda x: self.SYS_NAME in x.keys(),
                                      difference))
            curent_state = curent_state[0]
        return curent_state

    def show_change(self, difference, changes):
        """
        Метод вернет информацию об изменении в человекопонятном виде
        - перед изменениями покажет разницу с текущим состоянием БД
        - после изменения с состоянием на обработку сервиса
        :return:
            список из { имени поля, формат. старого значения,
                        формат. нового значения })
        :rtype: (bool, list)
        """
        verbose_name = self.model._meta.verbose_name
        rows = []
        curent_state = self._get_current_state(difference, changes)
        future_state = self._get_future_state(difference)

        new_list, delete_list = self._get_diff_data(
            curent_state, future_state)

        self._extend_show_result(
            lambda x: verbose_name,
            lambda x: '-',
            lambda x: x.display(),
            new_list,
            rows
        )
        self._extend_show_result(
            lambda x: verbose_name,
            lambda x: x.display(),
            lambda x: '-',
            delete_list,
            rows
        )
        return rows

    def _create_new(self, new_list, changes):
        """
        создаем привязку льготы к заявлению
        @param new_list:
        @return:
        """
        for priv_id in new_list:
            DeclarationPrivilege.objects.create(
                declaration=changes.declaration,
                privilege_id=priv_id
            )

    def apply_change(self, difference, changes):
        """
        применяит изменение в системе
        не возвращаем дикт с изменениями, тк он работает
        только для моделей заявки и ребенка
        @return:
        """
        curent_state = self._get_current_state(difference, changes)
        future_state = self._get_future_state(difference)
        new_list, delete_list = self._get_diff_data(
            curent_state, future_state)

        self._create_new(new_list, changes)
        for priv_id in delete_list:
            pr_qs = DeclarationPrivilege.objects.filter(
                declaration=changes.declaration,
                privilege_id=priv_id
            )
            if pr_qs.count() == 1:
                priv = pr_qs[0]
                priv.delete()
        return {}
